import random
from collections.abc import Callable
from datetime import datetime, UTC, timedelta
from typing import Optional

from pydantic import BaseModel

from massranker.elo import expected, elo

class ItemT(BaseModel):
    name: str
    elo: float
    enabled: bool


class StateT(BaseModel):
    items: list[ItemT]

    @classmethod
    def default(cls):
        return StateT(items=[])

    def _bring_up_to_date(self, names: list[str]):
        names_set = set(names)
        seen = set()

        # == first: figure out the list of enabled names ==
        for i in self.items:
            i.enabled = i.name in names_set
            seen.add(i.name)

        remaining_names = names_set - seen
        for name in remaining_names:
            self.items.append(ItemT(
                name=name,
                elo=1500,
                enabled=True,
            ))

    def do_tournament(self, names: list[str], callback: Callable[[ItemT, ItemT], Optional[ItemT]]):
        # bring up to date
        self._bring_up_to_date(names)

        # set up a tournament
        possible_entrants = [i for i in self.items if i.enabled]
        if len(possible_entrants) < 2:
            raise ValueError("need at least two things to compare")

        entrants = random.sample(possible_entrants, k=2)

        # ask for a winner
        winner = callback(entrants[0], entrants[1])

        # update elo
        score_0 = 0.0
        score_1 = 0.0

        if winner is None:
            score_0 += 0.5
            score_1 += 0.5
        elif winner == entrants[0]:
            score_0 += 1.0
        elif winner == entrants[1]:
            score_1 += 1.0

        expected_score_0 = expected(entrants[0].elo, entrants[1].elo)
        expected_score_1 = expected(entrants[1].elo, entrants[0].elo)
        # assert expected_score_0 == 1 - expected_score_1

        entrants[0].elo = elo(entrants[0].elo, expected_score_0, score_0)
        entrants[1].elo = elo(entrants[1].elo, expected_score_1, score_1)

    def dump_rankings(self) -> list[tuple[str, int]]:
        return [
            (e.name, int(e.elo))
            for e in sorted(self.items, key=lambda e: e.elo, reverse=True)
            if e.enabled
        ]